"""
Auto-generated by the update_vehicle_components_translation.py. Do not edit, ALL CHANGES WILL BE LOST.

This file is part of Ardupilot methodic configurator. https://github.com/ArduPilot/MethodicConfigurator

SPDX-FileCopyrightText: 2024-2025 Amilcar Lucas

SPDX-License-Identifier: GPL-3.0-or-later
"""

from ardupilot_methodic_configurator import _


def translatable_strings() -> None:
    """
    Translatable strings extracted from the all the vehicle_components.json files.

    For pygettext to extract them, they have no other function
    """
    _vehicle_components_strings = _("Battery")
    _vehicle_components_strings = _("Battery Monitor")
    _vehicle_components_strings = _("Capacity mAh")
    _vehicle_components_strings = _("Chemistry")
    _vehicle_components_strings = _("Components")
    _vehicle_components_strings = _("Configuration template")
    _vehicle_components_strings = _("Diameter_inches")
    _vehicle_components_strings = _("ESC")
    _vehicle_components_strings = _("FC Connection")
    _vehicle_components_strings = _("Firmware")
    _vehicle_components_strings = _("Flight Controller")
    _vehicle_components_strings = _("Format version")
    _vehicle_components_strings = _("Frame")
    _vehicle_components_strings = _("GNSS Receiver")
    _vehicle_components_strings = _("MCU Series")
    _vehicle_components_strings = _("Manufacturer")
    _vehicle_components_strings = _("Model")
    _vehicle_components_strings = _("Motors")
    _vehicle_components_strings = _("Notes")
    _vehicle_components_strings = _("Number of cells")
    _vehicle_components_strings = _("Poles")
    _vehicle_components_strings = _("Product")
    _vehicle_components_strings = _("Program version")
    _vehicle_components_strings = _("Propellers")
    _vehicle_components_strings = _("Protocol")
    _vehicle_components_strings = _("RC Controller")
    _vehicle_components_strings = _("RC Receiver")
    _vehicle_components_strings = _("RC Transmitter")
    _vehicle_components_strings = _("Specifications")
    _vehicle_components_strings = _("TOW max Kg")
    _vehicle_components_strings = _("TOW min Kg")
    _vehicle_components_strings = _("Telemetry")
    _vehicle_components_strings = _("Type")
    _vehicle_components_strings = _("URL")
    _vehicle_components_strings = _("Version")
    _vehicle_components_strings = _("Volt per cell crit")
    _vehicle_components_strings = _("Volt per cell low")
    _vehicle_components_strings = _("Volt per cell max")


def translatable_descriptions() -> None:  # noqa: PLR0915 # pylint: disable=too-many-statements
    """
    Translatable strings extracted from the vehicle_components_schema.json file.

    For pygettext to extract them, they have no other function
    """
    _vehicle_components_descriptions = _("Additional notes or comments about the component")
    _vehicle_components_descriptions = _("Base definition that all vehicle components inherit from")
    _vehicle_components_descriptions = _("Basic information about a product including manufacturer and model")
    _vehicle_components_descriptions = _("Battery capacity in milliampere-hours")
    _vehicle_components_descriptions = _("Battery chemistry type (e.g., LiPo, Li-Ion, LiFe)")
    _vehicle_components_descriptions = _("Collection of all components that make up the vehicle system")
    _vehicle_components_descriptions = _("Communication protocol used (e.g., SBUS, MAVLink)")
    _vehicle_components_descriptions = _("Company or brand that manufactures the component")
    _vehicle_components_descriptions = _("Component firmware information")
    _vehicle_components_descriptions = _("Component for wireless communication with ground station")
    _vehicle_components_descriptions = _("Component that connects to the flight controller")
    _vehicle_components_descriptions = _("Component that monitors battery voltage and current")
    _vehicle_components_descriptions = _("Connection type (e.g., UART, I2C, SPI)")
    _vehicle_components_descriptions = _("Critical voltage per cell below which damage may occur (e.g., 3.3V)")
    _vehicle_components_descriptions = _("Details about how a component connects to the flight controller")
    _vehicle_components_descriptions = _("Electronic Speed Controller for the motors")
    _vehicle_components_descriptions = _("Flight controller component that runs the ArduPilot firmware")
    _vehicle_components_descriptions = _("Flight controller firmware information")
    _vehicle_components_descriptions = _("Flight controller that runs ArduPilot firmware")
    _vehicle_components_descriptions = _("Global Navigation Satellite System receiver for positioning")
    _vehicle_components_descriptions = _("How this component connects to the flight controller")
    _vehicle_components_descriptions = _("Information about the firmware running on a component")
    _vehicle_components_descriptions = _("Low voltage threshold per cell when battery should be changed (e.g., 3.5V)")
    _vehicle_components_descriptions = _("Main power source for the vehicle")
    _vehicle_components_descriptions = _("Maximum take-off weight in kilograms")
    _vehicle_components_descriptions = _("Maximum voltage per cell when fully charged (e.g., 4.2V for LiPo)")
    _vehicle_components_descriptions = _("Microcontroller series used in the flight controller (e.g., STM32F4)")
    _vehicle_components_descriptions = _("Minimum take-off weight in kilograms")
    _vehicle_components_descriptions = _("Name of the configuration template")
    _vehicle_components_descriptions = _("Number of cells in series (e.g., 4S means 4 cells)")
    _vehicle_components_descriptions = _("Number of magnetic pole pairs in the motor (affects RPM calculation)")
    _vehicle_components_descriptions = _("Physical frame and structure of the vehicle")
    _vehicle_components_descriptions = _("Physical structure of the vehicle")
    _vehicle_components_descriptions = _("Power source for the vehicle")
    _vehicle_components_descriptions = _("Product information including manufacturer and model")
    _vehicle_components_descriptions = _("Propeller diameter in inches (affects PID tuning)")
    _vehicle_components_descriptions = _("Propellers that generate thrust")
    _vehicle_components_descriptions = _("Propellers that generate thrust for the vehicle")
    _vehicle_components_descriptions = _("Propulsion motors for the vehicle")
    _vehicle_components_descriptions = _("Propulsion motors that power the vehicle")
    _vehicle_components_descriptions = _("Receiver that connects to the flight controller")
    _vehicle_components_descriptions = _("Remote control component firmware information")
    _vehicle_components_descriptions = _("Remote control device used by the pilot")
    _vehicle_components_descriptions = _("Remote control system component")
    _vehicle_components_descriptions = _("Schema format version")
    _vehicle_components_descriptions = _("Specific model name/number of the component")
    _vehicle_components_descriptions = _("Technical specifications of the battery")
    _vehicle_components_descriptions = _("Technical specifications of the flight controller")
    _vehicle_components_descriptions = _("Technical specifications of the motors")
    _vehicle_components_descriptions = _("Technical specifications of the propellers")
    _vehicle_components_descriptions = _("Technical specifications of the vehicle frame")
    _vehicle_components_descriptions = _("Transmitter part of the remote control system")
    _vehicle_components_descriptions = _("Type of firmware")
    _vehicle_components_descriptions = _("Version number of the firmware")
    _vehicle_components_descriptions = _("Version of the program")
    _vehicle_components_descriptions = _("Version or revision of the hardware")
    _vehicle_components_descriptions = _("Web link to the product page or documentation")
